"""
The track module, that includes the
different types of tracks we accept for
the editor timeline.

A track is a sub-timeline in which we
can place the media that is accepted by
that track. The empty areas will be
fulfilled with imaginary 'empty' medias.
"""
# from yta_editor.track.audio import AudioTrack
# from yta_editor.track.video import VideoTrack


# __all__ = [
#     'AudioTrack',
#     'VideoTrack'
# ]